@@ -48,7 +48,7 @@ struct _GUPnPDeviceInfoPrivate {
 
         SoupURI *url_base;
 
-        XmlDocWrapper *doc;
+        GUPnPXMLDoc *doc;
 
         xmlNode *element;
 };
@@ -108,10 +108,7 @@ gupnp_device_info_set_property (GObject      *object,
 
                 break;
         case PROP_DOCUMENT:
-                info->priv->doc = g_value_get_object (value);
-                if (info->priv->doc)
-                        g_object_ref_sink (info->priv->doc);
-
+                info->priv->doc = g_value_dup_object (value);
                 break;
         case PROP_ELEMENT:
                 info->priv->element = g_value_get_pointer (value);
@@ -219,7 +216,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
         g_type_class_add_private (klass, sizeof (GUPnPDeviceInfoPrivate));
 
         /**
-         * GUPnPDeviceInfo:resource-factory
+         * GUPnPDeviceInfo:resource-factory:
          *
          * The resource factory to use. Set to NULL for default factory.
          **/
@@ -237,7 +234,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:context
+         * GUPnPDeviceInfo:context:
          *
          * The #GUPnPContext to use.
          **/
@@ -255,7 +252,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:location
+         * GUPnPDeviceInfo:location:
          *
          * The location of the device description file.
          **/
@@ -274,7 +271,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:udn
+         * GUPnPDeviceInfo:udn:
          *
          * The UDN of this device.
          **/
@@ -292,7 +289,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:device-type
+         * GUPnPDeviceInfo:device-type:
          *
          * The device type.
          **/
@@ -310,7 +307,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:url-base
+         * GUPnPDeviceInfo:url-base:
          *
          * The URL base (#SoupURI).
          **/
@@ -327,7 +324,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                        G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:document
+         * GUPnPDeviceInfo:document:
          *
          * Private property.
          *
@@ -340,7 +337,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       "Document",
                                       "The XML document related to this "
                                       "device",
-                                      TYPE_XML_DOC_WRAPPER,
+                                      GUPNP_TYPE_XML_DOC,
                                       G_PARAM_WRITABLE |
                                       G_PARAM_CONSTRUCT_ONLY |
                                       G_PARAM_STATIC_NAME |
@@ -348,7 +345,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPDeviceInfo:element
+         * GUPnPDeviceInfo:element:
          *
          * Private property.
          *
@@ -369,7 +366,7 @@ gupnp_device_info_class_init (GUPnPDeviceInfoClass *klass)
 }
 
 /**
- * gupnp_device_info_get_resource_factory
+ * gupnp_device_info_get_resource_factory:
  * @device_info: A #GUPnPDeviceInfo
  *
  * Get the #GUPnPResourceFactory used by the @device_info.
@@ -385,7 +382,7 @@ gupnp_device_info_get_resource_factory (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_context
+ * gupnp_device_info_get_context:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the associated #GUPnPContext.
@@ -401,7 +398,7 @@ gupnp_device_info_get_context (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_location
+ * gupnp_device_info_get_location:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the location of the device description file.
@@ -417,7 +414,7 @@ gupnp_device_info_get_location (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_url_base
+ * gupnp_device_info_get_url_base:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the URL base of this device.
@@ -433,7 +430,7 @@ gupnp_device_info_get_url_base (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_udn
+ * gupnp_device_info_get_udn:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the Unique Device Name of the device.
@@ -455,7 +452,7 @@ gupnp_device_info_get_udn (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_device_type
+ * gupnp_device_info_get_device_type:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the UPnP device type.
@@ -477,7 +474,7 @@ gupnp_device_info_get_device_type (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_friendly_name
+ * gupnp_device_info_get_friendly_name:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the friendly name of the device.
@@ -494,7 +491,7 @@ gupnp_device_info_get_friendly_name (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_manufacturer
+ * gupnp_device_info_get_manufacturer:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the manufacturer of the device.
@@ -511,7 +508,7 @@ gupnp_device_info_get_manufacturer (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_manufacturer_url
+ * gupnp_device_info_get_manufacturer_url:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a URL pointing to the manufacturer's website.
@@ -529,7 +526,7 @@ gupnp_device_info_get_manufacturer_url (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_model_description
+ * gupnp_device_info_get_model_description:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the description of the device model.
@@ -546,7 +543,7 @@ gupnp_device_info_get_model_description (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_model_name
+ * gupnp_device_info_get_model_name:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the model name of the device.
@@ -563,7 +560,7 @@ gupnp_device_info_get_model_name (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_model_number
+ * gupnp_device_info_get_model_number:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the model number of the device.
@@ -580,7 +577,7 @@ gupnp_device_info_get_model_number (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_model_url
+ * gupnp_device_info_get_model_url:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a URL pointing to the device model's website.
@@ -598,7 +595,7 @@ gupnp_device_info_get_model_url (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_serial_number
+ * gupnp_device_info_get_serial_number:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the serial number of the device.
@@ -615,7 +612,7 @@ gupnp_device_info_get_serial_number (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_upc
+ * gupnp_device_info_get_upc:
  * @info: A #GUPnPDeviceInfo
  *
  * Get the Universal Product Code of the device.
@@ -632,7 +629,7 @@ gupnp_device_info_get_upc (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_presentation_url
+ * gupnp_device_info_get_presentation_url:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a URL pointing to the device's presentation page, for web-based
@@ -694,19 +691,24 @@ icon_free (Icon *icon)
 }
 
 /**
- * gupnp_device_info_get_icon_url
+ * gupnp_device_info_get_icon_url:
  * @info: A #GUPnPDeviceInfo
- * @requested_mime_type: The requested file format, or %NULL for any
+ * @requested_mime_type: (allow-none) (transfer none): The requested file
+ * format, or %NULL for any
  * @requested_depth: The requested color depth, or -1 for any
  * @requested_width: The requested width, or -1 for any
  * @requested_height: The requested height, or -1 for any
  * @prefer_bigger: %TRUE if a bigger, rather than a smaller icon should be
  * returned if no exact match could be found
- * @mime_type: The location where to store the the format of the returned icon,
- * or %NULL. The returned string should be freed after use
- * @depth: The location where to store the depth of the returned icon, or %NULL
- * @width: The location where to store the width of the returned icon, or %NULL
- * @height: The location where to store the height of the returned icon, or %NULL
+ * @mime_type: (out) (allow-none): The location where to store the the format
+ * of the returned icon, or %NULL. The returned string should be freed after
+ * use
+ * @depth: (out) (allow-none) :  The location where to store the depth of the
+ * returned icon, or %NULL
+ * @width: (out) (allow-none) : The location where to store the width of the
+ * returned icon, or %NULL
+ * @height: (out) (allow-none) : The location where to store the height of the
+ * returned icon, or %NULL
  *
  * Get a URL pointing to the icon most closely matching the
  * given criteria, or %NULL. If @requested_mime_type is set, only icons with
@@ -716,7 +718,7 @@ icon_free (Icon *icon)
  * returned, unless @prefer_bigger is set, in which case the next biggest icon
  * will be returned. The returned strings should be freed.
  *
- * Return value: a string, or %NULL.  g_free() after use.
+ * Return value: (transfer full): a string, or %NULL.  g_free() after use.
  **/
 char *
 gupnp_device_info_get_icon_url (GUPnPDeviceInfo *info,
@@ -823,6 +825,8 @@ gupnp_device_info_get_icon_url (GUPnPDeviceInfo *info,
 
         /* Fill in return values */
         if (closest) {
+                icon = closest;
+
                 if (mime_type) {
                         if (icon->mime_type) {
                                 *mime_type = g_strdup
@@ -886,7 +890,7 @@ resource_type_match (const char *query,
 
         /* Inspect last colon (if any!) on @base */
         colon = strrchr (base, ':');
-        if (G_UNLIKELY (*colon == 0))
+        if (G_UNLIKELY (colon == NULL))
                 return !strcmp (query, base); /* No colon */
 
         /* Length of type until last colon */
@@ -926,7 +930,83 @@ resource_type_match (const char *query,
 }
 
 /**
- * gupnp_device_info_list_devices
+ * gupnp_device_info_list_dlna_capabilities:
+ * @info: A #GUPnPDeviceInfo
+ *
+ * Get a #GList of strings that represent the device capabilities as announced 
+ * in the device description file using the &lt;dlna:X_DLNACAP&gt; element.
+ *
+ * Return value: (element-type utf8): a #GList of newly allocated strings or
+ * %NULL if the device description doesn't contain the &lt;dlna:X_DLNACAP&gt;
+ * element.
+ **/
+GList *
+gupnp_device_info_list_dlna_capabilities (GUPnPDeviceInfo *info)
+{
+        xmlChar *caps;
+
+        g_return_val_if_fail (GUPNP_IS_DEVICE_INFO (info), NULL);
+
+        caps = xml_util_get_child_element_content (info->priv->element,
+                                                   "X_DLNACAP");
+
+        if (caps) {
+                GList         *list  = NULL;
+                const xmlChar *start = caps;
+
+                while (*start) {
+                        const xmlChar *end = start;
+
+                        while (*end && *end != ',')
+                                end++;
+
+                        if (end > start) {
+                                gchar *value;
+
+                                value = g_strndup ((const gchar *) start,
+                                                   end - start);
+
+                                list = g_list_prepend (list, value);
+                        }
+
+                        if (*end)
+                                start = end + 1;
+                        else
+                                break;
+                }
+
+                xmlFree (caps);
+
+                return g_list_reverse (list);
+        }
+
+        return NULL;
+}
+
+/**
+ * gupnp_device_info_get_description_value:
+ * @info:    A #GUPnPDeviceInfo
+ * @element: Name of the description element to retrieve
+ *
+ * This function provides generic access to the contents of arbitrary elements
+ * in the device description file.
+ *
+ * Return value: a newly allocated string or %NULL if the device
+ *               description doesn't contain the given @element
+ **/
+char *
+gupnp_device_info_get_description_value (GUPnPDeviceInfo *info,
+                                         const char      *element)
+{
+        g_return_val_if_fail (GUPNP_IS_DEVICE_INFO (info), NULL);
+        g_return_val_if_fail (element != NULL, NULL);
+
+        return xml_util_get_child_element_content_glib (info->priv->element,
+                                                        element);
+}
+
+/**
+ * gupnp_device_info_list_devices:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a #GList of new objects implementing #GUPnPDeviceInfo
@@ -938,7 +1018,8 @@ resource_type_match (const char *query,
  * must cache any used devices if it wishes to keep them around and re-use
  * them.
  *
- * Return value: a #GList of new #GUPnPDeviceInfo objects.
+ * Return value: (element-type GUPnP.DeviceInfo) (transfer full): a #GList of
+ * new #GUPnPDeviceInfo objects.
  **/
 GList *
 gupnp_device_info_list_devices (GUPnPDeviceInfo *info)
@@ -974,14 +1055,14 @@ gupnp_device_info_list_devices (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_list_device_types
+ * gupnp_device_info_list_device_types:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a #GList of strings representing the types of the devices
  * directly contained in @info.
  *
- * Return value: A #GList of strings. The elements should be g_free()'d and the
- * list should be g_list_free()'d.
+ * Return value: (element-type utf8) (transfer full): A #GList of strings. The
+ * elements should be g_free()'d and the list should be g_list_free()'d.
  **/
 GList *
 gupnp_device_info_list_device_types (GUPnPDeviceInfo *info)
@@ -1016,7 +1097,7 @@ gupnp_device_info_list_device_types (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_device
+ * gupnp_device_info_get_device:
  * @info: A #GUPnPDeviceInfo
  * @type: The type of the device to be retrieved.
  *
@@ -1024,10 +1105,9 @@ gupnp_device_info_list_device_types (GUPnPDeviceInfo *info)
  * a new object implementing #GUPnPDeviceInfo, or %NULL if no such device
  * was found. The returned object should be unreffed when done.
  *
- * Note that devices are not cached internally, so that every time you
- * call this function a new object is created. The application
- * must cache any used devices if it wishes to keep them around and re-use
- * them.
+ * Note that devices are not cached internally, so that every time you call
+ * this function a new object is created. The application must cache any used
+ * devices if it wishes to keep them around and re-use them.
  *
  * Return value: A new #GUPnPDeviceInfo.
  **/
@@ -1080,22 +1160,21 @@ gupnp_device_info_get_device (GUPnPDeviceInfo *info,
 
         return device;
 }
-
 /**
- * gupnp_device_info_list_services
+ * gupnp_device_info_list_services:
  * @info: A #GUPnPDeviceInfo
  *
- * Get a #GList of new objects implementing #GUPnPServiceInfo
- * representing the services directly contained in @info. The returned list
- * should be g_list_free()'d and the elements should be g_object_unref()'d.
+ * Get a #GList of new objects implementing #GUPnPServiceInfo representing the
+ * services directly contained in @info. The returned list should be
+ * g_list_free()'d and the elements should be g_object_unref()'d.
  *
- * Note that services are not cached internally, so that every time you
- * call this function new objects are created. The application
- * must cache any used services if it wishes to keep them around and re-use
- * them.
+ * Note that services are not cached internally, so that every time you call
+ * function new objects are created. The application must cache any used
+ * services if it wishes to keep them around and re-use them.
  *
- * Return value: A #GList of new #GUPnPServiceInfo objects.
- **/
+ * Return value: (element-type GUPnP.ServiceInfo) (transfer full) : A #GList of
+ * new #GUPnPServiceInfo objects.
+ */
 GList *
 gupnp_device_info_list_services (GUPnPDeviceInfo *info)
 {
@@ -1130,14 +1209,14 @@ gupnp_device_info_list_services (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_list_service_types
+ * gupnp_device_info_list_service_types:
  * @info: A #GUPnPDeviceInfo
  *
  * Get a #GList of strings representing the types of the services
  * directly contained in @info.
  *
- * Return value: A #GList of strings. The elements should be g_free()'d and the
- * list should be g_list_free()'d.
+ * Return value: (element-type utf8) (transfer full): A #GList of strings. The
+ * elements should be g_free()'d and the list should be g_list_free()'d.
  **/
 GList *
 gupnp_device_info_list_service_types (GUPnPDeviceInfo *info)
@@ -1172,18 +1251,17 @@ gupnp_device_info_list_service_types (GUPnPDeviceInfo *info)
 }
 
 /**
- * gupnp_device_info_get_service
+ * gupnp_device_info_get_service:
  * @info: A #GUPnPDeviceInfo
  * @type: The type of the service to be retrieved.
  *
- * Get the service with type @type directly contained in @info as
- * a new object implementing #GUPnPServiceInfo, or %NULL if no such device
- * was found. The returned object should be unreffed when done.
+ * Get the service with type @type directly contained in @info as a new object
+ * implementing #GUPnPServiceInfo, or %NULL if no such device was found. The
+ * returned object should be unreffed when done.
  *
- * Note that services are not cached internally, so that every time you
- * call this function a new object is created. The application
- * must cache any used services if it wishes to keep them around and re-use
- * them.
+ * Note that services are not cached internally, so that every time you call
+ * this function a new object is created. The application must cache any used
+ * services if it wishes to keep them around and re-use them.
  *
  * Return value: A #GUPnPServiceInfo.
  **/
@@ -1237,8 +1315,8 @@ gupnp_device_info_get_service (GUPnPDeviceInfo *info,
         return service;
 }
 
-/* Return associated xmlDoc wrapper */
-XmlDocWrapper *
+/* Return associated xmlDoc */
+GUPnPXMLDoc *
 _gupnp_device_info_get_document (GUPnPDeviceInfo *info)
 {
         return info->priv->doc;
