@@ -874,7 +874,6 @@ control_server_handler (SoupServer        *server,
         char *action_name;
         char *end;
         GUPnPServiceAction *action;
-        goffset length;
 
         service = GUPNP_SERVICE (user_data);
 
@@ -884,8 +883,7 @@ control_server_handler (SoupServer        *server,
                 return;
         }
 
-        length = soup_message_headers_get_content_length (msg->request_headers);
-        if (length == 0) {
+        if (msg->request_body->length == 0) {
                 soup_message_set_status (msg, SOUP_STATUS_BAD_REQUEST);
 
                 return;
